use serde::{Deserialize, Serialize};

use domain::base::{iana::Class, Name, Record as DnsRecord, Ttl};

use crate::{errors::Error, validation};
use crate::macros::{append_errors, push_error};

use crate::resources::dns::external;
use crate::resources::dns::internal;
use crate::proto;

pub enum RecordParseError {
    Ip4Address { input: String },
    Ip6Address { input: String },
    RDataUnknown { input: String, field: String, rtype: String },
    NameUnknown { input: String },
    NotInZone { name: String, zone: String },
}

pub enum RecordError {
    Validation { suberrors: Vec<Error> },
}

pub(crate) type DnsRecordImpl = DnsRecord<
    Name<Vec<u8>>,
    proto::dns::ParsedRData<Name<Vec<u8>>,Vec<u8>>
>;

#[derive(Debug, Deserialize, Serialize)]
pub struct Record {
    pub name: String,
    pub ttl: u32,
    #[serde(flatten)]
    pub rdata: external::RData
}

// TODO: Proto
impl<Name: ToString, Oct: AsRef<[u8]>> From<DnsRecord<Name, proto::dns::ParsedRData<Name, Oct>>> for Record {
    fn from(value: DnsRecord<Name, proto::dns::ParsedRData<Name, Oct>>) -> Self {
        Record {
            name: value.owner().to_string(),
            ttl: value.ttl().as_secs(),
            rdata: internal::RData::from(value.into_data()).into(),
        }
    }
}

impl Record {
    fn convert(self, zone_name: &Name<Vec<u8>>) -> Result<DnsRecordImpl, Vec<Error>> {
        let mut errors = Vec::new();

        let name = push_error!(validation::normalize_domain(&self.name), errors, "/name");

        let name = name.and_then(|name| push_error!(name.parse::<Name<_>>().map_err(|e| {
            Error::from(RecordParseError::NameUnknown {
                input: self.name.clone()
            }).with_cause(&e.to_string())
        }), errors, "/name"));

        let name = name.and_then(|name| {
            if !name.ends_with(zone_name) {
                errors.push(
                    Error::from(RecordParseError::NotInZone { name: self.name, zone: zone_name.to_string() })
                        .with_path("/name")
                );
                None
            } else {
                Some(name)
            }
        });

        let ttl = Ttl::from_secs(self.ttl);
        let rdata = append_errors!(self.rdata.validate(), errors, "/rdata");


        if errors.is_empty() {
            // TODO: Split this in proto / external
            let rdata = proto::dns::ParsedRData::try_from(rdata.unwrap()).unwrap();
            Ok(DnsRecord::new(name.unwrap(), Class::IN, ttl, rdata))
        } else {
            Err(errors)
        }
    }
}


#[derive(Debug, Deserialize, Serialize)]
pub struct RecordList(pub Vec<Record>);

impl RecordList {
    fn convert(self, zone_name: &Name<Vec<u8>>) -> Result<Vec<DnsRecordImpl>, Vec<Error>> {
        let mut errors = Vec::new();
        let mut records = Vec::new();

        for (index, record) in self.0.into_iter().enumerate() {
            let record = append_errors!(record.convert(zone_name), errors, &format!("/{index}"));

            if let Some(record) = record {
                records.push(record)
            }
        }

        if errors.is_empty() {
            Ok(records)
        } else {
            Err(errors)
        }
    }
}

#[derive(Debug,Deserialize)]
pub struct AddRecordsRequest {
    pub new_records: RecordList
}

pub struct AddRecords {
    pub new_records: Vec<DnsRecordImpl>
}

impl AddRecordsRequest {
    pub fn validate(self, zone_name: &str) -> Result<AddRecords, Error> {
        let zone_name: Name<Vec<u8>> = zone_name.parse().expect("zone name is assumed to be valid");

        let mut errors = Vec::new();
        let records = append_errors!(self.new_records.convert(&zone_name), errors, "/new_records");

        if errors.is_empty() {
            Ok(AddRecords {
                new_records: records.unwrap(),
            })
        } else {
            Err(Error::from(RecordError::Validation { suberrors: errors }))
        }
    }
}
